/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/







#include "dcGainCont.h"

ostream& operator<<
(ostream& os, const DestCentricGainContainerT& gc)
{
    if(gc.getMaxGain() == -1)
	cout<<"setMaxGain has not been called. "
	    <<"Can't print DestCentricGainContainerT";

    os<<"  MaxGain:  "<<gc.getMaxGain()<<endl;

    for(unsigned k = 0; k < gc.getNumPrioritizers(); k++)
    {
        os<<"Moves To Partition "<<k<<endl;

        for(int g = gc.getPrioritizer(k).size()-1; g >= 0; g--)
        {
            os << setw(4) << g - gc.getMaxGain() <<") ";
            SVGainElement* cur =
                static_cast<SVGainElement* const>(gc.getPrioritizer(k)[g]._next);
            if(cur == NULL)
                os<<" EMPTY "<<endl;
            else
            {
                for( ; cur != NULL;
                        cur = static_cast<SVGainElement* const>(cur->_next))
                {

                    os<<"[#:"<<gc.getRepository().getElementId(*cur);
                    os<<" g:"<<cur->getGain()<<" + ";
                    os<<cur->getRealGain() - cur->getGain()<<"]";

                }
                os<<endl;
            }
        }
    }
    return os;
}

void DestCentricGainContainerT::reinitialize()
{
    for(unsigned k = 0; k < _prioritizer.size(); k++)
        _prioritizer[k].removeAll();

    _highestArray = 0;
}

void DestCentricGainContainerT::setupClip()
{
    for(unsigned k = 0; k < _prioritizer.size(); k++)
	_prioritizer[k].setupForClip();
}

void DestCentricGainContainerT::setMaxGain(unsigned maxG)
{
    _maxGain = maxG;
    for(unsigned k = 0; k < _prioritizer.size(); k++)
        _prioritizer[k].setMaxGain(maxG);
}

bool DestCentricGainContainerT::isConsistent()
{
    bool consistent = true;
    for(unsigned k = 0; k < _prioritizer.size(); k++)
    {
        if(! _prioritizer[k].isConsistent())
        {
            consistent = false;
            cout<<"BucketArray "<<k<<" is not consistent"<<endl;
        }
    }

    return consistent;
}
